/******************************************************************
Copyright 2000 by Object Craft P/L, Melbourne, Australia.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Object Craft
is not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

OBJECT CRAFT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL OBJECT CRAFT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

#include <ctype.h>
#include "Python.h"
#include "structmember.h"

static PyObject *error_obj;	/* CSV exception */

/* ----------------------------------------------------- */

/* Declarations for objects of type Parser */

typedef enum {
    START_RECORD, START_FIELD,
    IN_FIELD, QUOTE_IN_FIELD,
    IN_QUOTED_FIELD, QUOTE_IN_QUOTED_FIELD,
    END_QUOTED_FIELD
} ParserState;

typedef struct {
    PyObject_HEAD

    int ms_double_quote;	/* is " represented by ""? */
    char field_sep;		/* field separator */

    ParserState state;		/* current CSV parse state */
    PyObject *fields;		/* field list for current record */

    int auto_clear;		/* should fields be cleared on next
                                   parse() after exception? */
    int had_parse_error;	/* did we have a parse error? */

    char *field;		/* build current field in here */
    int field_size;		/* size of allocated buffer */
    int field_len;		/* length of current field */

    char *rec;			/* buffer for parser.join */
    int rec_size;		/* size of allocated record */
    int rec_len;		/* length of record */
    int num_fields;		/* number of fields in record */
} ParserObj;

staticforward PyTypeObject ParserType;

static PyObject *raise_exception(char *fmt, ...)
{
    va_list ap;
    char msg[512];
    PyObject *pymsg;

    va_start(ap, fmt);
#ifdef _WIN32
    _vsnprintf(msg, sizeof(msg), fmt, ap);
#else
    vsnprintf(msg, sizeof(msg), fmt, ap);
#endif
    va_end(ap);
    pymsg = PyString_FromString(msg);
    PyErr_SetObject(error_obj, pymsg);
    Py_XDECREF(pymsg);

    return NULL;
}

static void parse_save_field(ParserObj *self)
{
    PyObject *field;

    field = PyString_FromStringAndSize(self->field, self->field_len);
    if (field != NULL) {
	PyList_Append(self->fields, field);
	Py_XDECREF(field);
    }
    self->field_len = 0;
}

static void parse_add_char(ParserObj *self, char c)
{
    if (self->field_len == self->field_size) {
	if (self->field_size == 0) {
	    self->field_size = 4096;
	    self->field = malloc(self->field_size);
	} else {
	    self->field_size *= 2;
	    self->field = realloc(self->field, self->field_size);
	}
	if (self->field == NULL) {
	    PyErr_NoMemory();
	    return;
	}
    }
    self->field[self->field_len++] = c;
}

static void parse_process_char(ParserObj *self, char c)
{
    switch (self->state) {
    case START_RECORD:
	/* start of record */
	if (c == '\0')
	    /* empty line - return [] */
	    break;
	/* normal character - handle as START_FIELD */
	self->state = START_FIELD;
    case START_FIELD:
	/* expecting field */
	if (c == '"') {
	    /* start quoted field */
	    self->state = IN_QUOTED_FIELD;
	} else if (c == self->field_sep) {
	    /* save empty field */
	    parse_save_field(self);
	} else if (c == '\0') {
	    /* save empty field - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	} else {
	    /* begin new unquoted field */
	    parse_add_char(self, c);
	    self->state = IN_FIELD;
	}
	break;

    case IN_FIELD:
	/* in unquoted field */
	if (c == self->field_sep) {
	    /* save field - wait for new field */
	    parse_save_field(self);
	    self->state = START_FIELD;
	} else if (c == '\0') {
	    /* end of line - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	} else if (c == '"' && self->ms_double_quote) {
	    /* microsoft style double quotes; " represented by "" */
	    self->state = QUOTE_IN_FIELD;
	} else {
	    /* normal character - save in field */
	    parse_add_char(self, c);
	}
	break;

    case QUOTE_IN_FIELD:
	/* microsoft double quotes - seen a quote in an unquoted field */
	parse_add_char(self, '"');
	if (c == '"') {
	    /* save "" as " */
	    self->state = IN_FIELD;
	} else if (c == '\0') {
	    /* end of line - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	} else if (c == self->field_sep) {
	    /* save field - wait for new field */
	    parse_save_field(self);
	    self->state = START_FIELD;
	} else {
	    /* normal character - save in field */
	    parse_add_char(self, c);
	    self->state = IN_FIELD;
	}
	break;

    case IN_QUOTED_FIELD:
	/* in quoted field */
	if (c == '"') {
	    if (self->ms_double_quote) {
		/* microsoft style double quotes; " represented by "" */
		self->state = QUOTE_IN_QUOTED_FIELD;
	    } else {
		/* end of field */
		parse_save_field(self);
		self->state = END_QUOTED_FIELD;
	    }
	} else if (c == '\0') {
	    /* end of line - save '\n' in field */
	    parse_add_char(self, '\n');
	} else {
	    /* normal character - save in field */
	    parse_add_char(self, c);
	}
	break;

    case QUOTE_IN_QUOTED_FIELD:
	/* microsoft double quotes - seen a quote in an quoted field */
	if (c == '"') {
	    /* save "" as " */
	    parse_add_char(self, '"');
	    self->state = IN_QUOTED_FIELD;
	} else if (c == self->field_sep) {
	    /* save field - wait for new field */
	    parse_save_field(self);
	    self->state = START_FIELD;
	} else if (c == '\0') {
	    /* end of line - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	} else {
	    /* illegal */
	    self->had_parse_error = 1;
	    raise_exception("%c expected after \"", self->field_sep);
	}
	break;

    case END_QUOTED_FIELD:
	/* seen closing " on quoted field */
	if (c == self->field_sep) {
	    /* wait for new field */
	    self->state = START_FIELD;
	} else if (c == '\0') {
	    /* end of line - return [fields] */
	    self->state = START_RECORD;
	} else {
	    self->had_parse_error = 1;
	    raise_exception("%c expected after \"", self->field_sep);
	}
	break;
    }
}

static void clear_fields_and_status(ParserObj *self)
{
    if (self->fields) {
	Py_XDECREF(self->fields);
    }
    self->fields = PyList_New(0);
    self->field_len = 0;
    self->state = START_RECORD;

    self->had_parse_error = 0;
}

/* ---------------------------------------------------------------- */

static char Parser_parse__doc__[] = 
"parse(s) -> list of strings\n"
"\n"
"CSV parse the single line in the string s and return a\n"
"list of string fields.  If the CSV record contains multi-line\n"
"fields, the function will return None until all lines of the\n"
"record have been parsed.";

static PyObject *Parser_parse(ParserObj *self, PyObject *args)
{
    char *line;

    if (!PyArg_ParseTuple(args, "s", &line))
	return NULL;

    if (self->auto_clear && self->had_parse_error)
	clear_fields_and_status(self);

    /* Process line of text - send '\0' to processing code to
       represent end of line.  End of line which is not at end of
       string is an error. */
    while (*line) {
	char c;

	c = *line++;
	if (c == '\r') {
	    c = *line++;
	    if (c == '\0')
		/* macintosh end of line */
		break;
	    if (c == '\n') {
		c = *line++;
		if (c == '\0')
		    /* DOS end of line */
		    break;
	    }
	    self->had_parse_error = 1;
	    return raise_exception("Newline inside string");
	}
	if (c == '\n') {
	    c = *line++;
	    if (c == '\0')
		/* unix end of line */
		break;
	    self->had_parse_error = 1;
	    return raise_exception("Newline inside string");
	}
	parse_process_char(self, c);
	if (PyErr_Occurred())
	    return NULL;
    }
    parse_process_char(self, '\0');

    if (self->state == START_RECORD) {
	PyObject *fields = self->fields;
	self->fields = PyList_New(0);
	return fields;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

/* ---------------------------------------------------------------- */

static char Parser_clear__doc__[] = 
"clear() -> None\n"
"\n"
"Discard partially parsed record.  This must be called to reset\n"
"parser state after an exception.";

static PyObject *Parser_clear(ParserObj *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    clear_fields_and_status(self);

    Py_INCREF(Py_None);
    return Py_None;
}

/* ---------------------------------------------------------------- */
static void join_reset(ParserObj *self)
{
    self->rec_len = 0;
    self->num_fields = 0;
}

#define MEM_INCR 32768

static int join_append(ParserObj *self, char *field, int quote_empty)
{
    int len, i, quoted;

    /* work out field length */
    for (i = len = quoted = 0;; i++, len++) {
	char c = field[i];
	if (c == '\0')
	    break;
	if (c == '"') {
	    quoted = 1;
	    len++;
	} else if (isspace((int)c) || c == self->field_sep)
	    quoted = 1;
    }
    if (quote_empty && len == 0)
	quoted = 1;
    if (quoted)
	len += 2;		/* field is quoted */
    if (self->num_fields > 0)
	len++;			/* need comma before field */

    /* grow record buffer if necessary */
    if (self->rec_len + len > self->rec_size) {
	if (self->rec_size == 0) {
	    self->rec_size = (len / MEM_INCR + 1) * MEM_INCR;
	    self->rec = malloc(self->rec_size);
	} else {
	    char *old_rec = self->rec;

	    self->rec_size =
		((self->rec_size + len) / MEM_INCR + 1) * MEM_INCR;
	    self->rec = realloc(self->rec, self->rec_size);
	    if (self->rec == NULL)
		free(old_rec);
	}
	if (self->rec == NULL) {
	    PyErr_NoMemory();
	    return 0;
	}
    }

    /* copy field into record buffer */
    if (self->num_fields > 0)
	self->rec[self->rec_len++] = self->field_sep;
    if (quoted)
	self->rec[self->rec_len++] = '"';
    for (i = 0;; i++) {
	char c = field[i];
	if (c == '\0')
	    break;
	if (c == '"')
	    self->rec[self->rec_len++] = '"';
	self->rec[self->rec_len++] = c;
    }
    if (quoted)
	self->rec[self->rec_len++] = '"';
    self->num_fields++;

    return 1;
}

static PyObject *join_string(ParserObj *self)
{
    return PyString_FromStringAndSize(self->rec, self->rec_len);
}

static char Parser_join__doc__[] = 
"join(sequence) -> string\n"
"\n"
"Construct a CSV record from a sequence of fields.  Non-string\n"
"elements will be converted to string.";

static PyObject *Parser_join(ParserObj *self, PyObject *args)
{
    PyObject *seq;
    int len, i;

    if (!PyArg_ParseTuple(args, "O", &seq))
	return NULL;
    if (!PySequence_Check(seq))
	return raise_exception("sequence expected");

    len = PySequence_Length(seq);
    if (len < 0)
	return NULL;

    join_reset(self);
    for (i = 0; i < len; i++) {
	PyObject *field;
	int append_ok;

	field = PySequence_GetItem(seq, i);
	if (field == NULL)
	    return NULL;

	if (PyString_Check(field)) {
	    append_ok = join_append(self, PyString_AsString(field), len == 1);
	    Py_DECREF(field);
	} else {
	    PyObject *str;

	    str = PyObject_Str(field);
	    Py_DECREF(field);
	    if (str == NULL)
		return NULL;

	    append_ok = join_append(self, PyString_AsString(str), len == 1);
	    Py_DECREF(str);
	}
	if (!append_ok)
	    return NULL;
    }

    return join_string(self);
}

static struct PyMethodDef Parser_methods[] = {
    { "parse", (PyCFunction)Parser_parse, METH_VARARGS, Parser_parse__doc__ },
    { "clear", (PyCFunction)Parser_clear, METH_VARARGS, Parser_clear__doc__ },
    { "join", (PyCFunction)Parser_join, METH_VARARGS, Parser_join__doc__ },
    
    { NULL, NULL }		/* sentinel */
};

static void Parser_dealloc(ParserObj *self)
{
    if (self->field)
	free(self->field);
    if (self->fields) {
	Py_XDECREF(self->fields);
    }
    if (self->rec)
	free(self->rec);

    PyMem_DEL(self);
}

#define OFF(x) offsetof(ParserObj, x)

static struct memberlist Parser_memberlist[] = {
    { "ms_double_quote", T_INT,    OFF(ms_double_quote) },
    { "fields",          T_OBJECT, OFF(fields) },
    { "field_sep",       T_CHAR,   OFF(field_sep) },
    { "auto_clear",      T_INT,    OFF(auto_clear) },
    { "had_parse_error", T_INT,    OFF(had_parse_error), RO },

    { NULL }  /* Sentinel */
};

static PyObject *Parser_getattr(ParserObj *self, char *name)
{
    PyObject *rv;

    rv = PyMember_Get((char *)self, Parser_memberlist, name);
    if (rv)
	return rv;
    PyErr_Clear();
    return Py_FindMethod(Parser_methods, (PyObject *)self, name);
}

static int Parser_setattr(ParserObj *self, char *name, PyObject *v)
{
    if (v == NULL) {
	PyErr_SetString(PyExc_AttributeError, "Cannot delete attribute");
	return -1;
    }
    /* can remove this after Python setattr T_CHAR bug is fixed */
    if (strcmp(name, "field_sep") == 0) {
	if (PyString_Check(v) && PyString_Size(v) == 1) {
	    self->field_sep = PyString_AsString(v)[0];
	    return 0;
	} else {
	    PyErr_BadArgument();
	    return -1;
	}
    }
    return PyMember_Set((char *)self, Parser_memberlist, name, v);
}

static char ParserType__doc__[] = 
"CSV parser";

static PyTypeObject ParserType = {
    PyObject_HEAD_INIT(0)
    0,				/*ob_size*/
    "Parser",			/*tp_name*/
    sizeof(ParserObj),		/*tp_basicsize*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)Parser_dealloc,	/*tp_dealloc*/
    (printfunc)0,		/*tp_print*/
    (getattrfunc)Parser_getattr, /*tp_getattr*/
    (setattrfunc)Parser_setattr, /*tp_setattr*/
    (cmpfunc)0,			/*tp_compare*/
    (reprfunc)0,		/*tp_repr*/
    0,				/*tp_as_number*/
    0,				/*tp_as_sequence*/
    0,				/*tp_as_mapping*/
    (hashfunc)0,		/*tp_hash*/
    (ternaryfunc)0,		/*tp_call*/
    (reprfunc)0,		/*tp_str*/

    /* Space for future expansion */
    0L, 0L, 0L, 0L,
    ParserType__doc__		/* Documentation string */
};

/* End of code for Parser objects */
/* -------------------------------------------------------- */

static char csv_parser__doc__[] =
"parser(ms_double_quote = 1, field_sep = ',', auto_clear = 1) -> Parser\n"
"\n"
"Constructs a CSV parser object.  When ms_double_quote is TRUE\n"
"a double quotes in a fields must appear as \"\".\n"
"field_sep defines the character that will be used to separate\n"
"fields in the CSV record.\n"
"When auto_clear is TRUE calling parse() will automatically call\n"
"the clear() method if the previous call to parse() raised an\n"
"exception during parsing.";

static PyObject *csv_parser(PyObject *module,
			    PyObject *args, PyObject *keyword_args)
{
    static char *keywords[] = {
	"ms_double_quote", "field_sep", "auto_clear", NULL
    };
    ParserObj *self = PyObject_NEW(ParserObj, &ParserType);

    if (self == NULL)
	return NULL;
    self->state = START_RECORD;
    self->fields = PyList_New(0);
    self->ms_double_quote = 1;
    self->field_sep = ',';
    self->auto_clear = 1;

    self->had_parse_error = 0;

    self->field = NULL;
    self->field_size = 0;
    self->field_len = 0;

    self->rec = NULL;
    self->rec_size = 0;
    self->rec_len = 0;
    self->num_fields = 0;

    if (PyArg_ParseTupleAndKeywords(args, keyword_args, "|ici", keywords,
				    &self->ms_double_quote, &self->field_sep,
				    &self->auto_clear))
	return (PyObject*)self;

    Py_DECREF(self);
    return NULL;
}

/* List of methods defined in the module */

static struct PyMethodDef csv_methods[] = {
    { "parser", (PyCFunction)csv_parser, METH_VARARGS | METH_KEYWORDS, csv_parser__doc__ },

    { NULL, (PyCFunction)NULL, 0, NULL } /* sentinel */
};

/* Initialization function for the module (*must* be called initcsv) */

static char csv_module__doc__[] = 
"CSV parser / writer.";

void initcsv(void)
{
    PyObject *module, *dict, *ver = NULL;

    ParserType.ob_type = &PyType_Type;
    /* Create the module and add the functions */
    module = Py_InitModule4("csv", csv_methods,
			    csv_module__doc__,
			    (PyObject*)NULL,PYTHON_API_VERSION);
    if (module == NULL)
	goto error;

    /* Add some symbolic constants to the module */
    if ((dict = PyModule_GetDict(module)) == NULL)
	goto error;
    if ((ver = PyString_FromString("0.5")) == NULL)
	goto error;
    if (PyDict_SetItemString(dict, "__version__", ver) < 0)
	goto error;
    if ((error_obj = PyErr_NewException("csv.Error", NULL, NULL)) == NULL)
	goto error;
    PyDict_SetItemString(dict, "Error", error_obj);

error:
    Py_XDECREF(ver);
    Py_XDECREF(error_obj);
    /* Check for errors */
    if (PyErr_Occurred())
	Py_FatalError("can't initialize module csv");
}
